import datetime
import re
from typing import Iterable

float_pattern = r"[-+]?((\d*\.\d+)|(\d+\.?))([Ee][+-]?\d+)?"

int_regex = re.compile(r"[-+]?\d+$")
float_regex = re.compile(r"[-+]?((\d*\.\d+)|(\d+\.?))([Ee][+-]?\d+)?$")
percentage_regex = re.compile(r"[-+]?((\d*\.\d+)|(\d+\.?))([Ee][+-]?\d+)?%$")
date_time_regex = re.compile(r"\D+\s\D+\s\d+\s\d+:\d+:\d+\s\d{4}")


def convert_data_types(value):
    """Convert the given value string to the type it matches."""
    if value is None or value == "-":
        # Commonly used sentinel for a missing value in log tables
        return None
    elif int_regex.match(value):
        return int(value)
    elif float_regex.match(value):
        return float(value)
    elif percentage_regex.match(value):
        return float(value.rstrip("%")) / 100
    elif date_time_regex.match(value):
        return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Y")
    else:
        return value


def typeconvert_groupdict(match: re.Match):
    """Return the groupdict of a regex match with type converted values."""
    return {k: convert_data_types(v) for k, v in match.groupdict().items()}


def parse_lines(parser, loglines: Iterable[str]):
    """Parse the given lines using the given parser object.

    This function is mainly used in the tests.
    """
    lines = iter(loglines)
    for line in lines:
        parser.parse(line)


def parse_block(parser, log):
    """Parse a multi-line block of text using the given parser object.

    This function is mainly used in the tests.
    """
    parse_lines(parser, log.strip().split("\n"))


def model_type(discrete_vars=0, quad_nonzeros=0, quad_constrs=0):
    """Return the type of the optimization model.

    The model type is MIP, QCP, QP, or LP.
    """
    mtype = ""
    if discrete_vars > 0:
        mtype += "MI"
    if quad_constrs > 0:
        mtype += "QC"
    elif quad_nonzeros > 0:
        mtype += "Q"
    if not mtype:
        return "LP"
    return mtype + "P"



import re

from grblogtools.parsers.util import typeconvert_groupdict


class NoRelParser:
    norel_log_start = re.compile(r"Starting NoRel heuristic")
    norel_primal_regex = re.compile(
        r"Found heuristic solution:\sobjective\s(?P<Incumbent>[^\s]+)"
    )
    # Order is important in this list as regexes are checked in order
    norel_elapsed = [
        re.compile(
            r"Elapsed time for NoRel heuristic:\s(?P<Time>\d+)s\s\(best\sbound\s(?P<BestBd>[^\s]+)\)"
        ),
        re.compile(r"Elapsed time for NoRel heuristic:\s(?P<Time>\d+)s"),
    ]

    def __init__(self):
        self._progress = []
        self._incumbent = None
        self._started = False

    def get_summary(self) -> dict:
        """Return the summary based on the timeline information.

        It assumes that the best bound is always found in the last line, if exists.
        """
        if not self._progress:
            return {}
        last_log = self._progress[-1]
        result = {"NoRelTime": last_log["Time"]}
        if "BestBd" in last_log:
            result["NoRelBestBd"] = last_log["BestBd"]
        if self._incumbent is not None:
            result["NoRelBestSol"] = self._incumbent
        return result

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary and progress data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        if not self._started:
            match = self.norel_log_start.match(line)
            if match:
                self._started = True
                return True
            return False

        match = self.norel_primal_regex.match(line)
        if match:
            self._incumbent = float(match.group("Incumbent"))
            return True

        for regex in self.norel_elapsed:
            match = regex.match(line)
            if match:
                entry = typeconvert_groupdict(match)
                if self._incumbent is not None:
                    entry["Incumbent"] = self._incumbent
                self._progress.append(entry)
                return True

        return False

    def get_progress(self) -> list:
        """Return the progress of the norel heuristic."""
        return self._progress



import json
from functools import lru_cache
from pathlib import Path

data_dir = Path(__file__).parent.joinpath("data")


@lru_cache()
def load_descriptions():
    with data_dir.joinpath("descriptions.json").open() as infile:
        descriptions = json.load(infile)
    return {
        f"{parameter} (Parameter)": {
            entry["value"]: "{value}: {description}".format(**entry)
            for entry in entries
        }
        for parameter, entries in descriptions.items()
    }



import re

from grblogtools.parsers.util import typeconvert_groupdict


class PresolveParser:
    # The pattern indicating the initialization of the parser
    presolve_start_pattern = re.compile(
        r"Optimize a model with (?P<NumConstrs>\d+) (R|r)ows, (?P<NumVars>\d+) (C|c)olumns and (?P<NumNZs>\d+) (N|n)on(Z|z)ero(e?)s"
    )

    # Possible intermediate patterns to be parsed
    presolve_intermediate_patterns = [
        re.compile(r"Model fingerprint: (?P<Fingerprint>.*)$"),
        re.compile(
            r"Variable types: (?P<PresolvedNumConVars>\d+) continuous, (?P<PresolvedNumIntVars>\d+) integer \((?P<PresolvedNumBinVars>\d+) binary\)$"
        ),
        re.compile(
            r"Variable types: (?P<PresolvedNumBinVars>\d+) bin/(?P<PresolvedNumIntVars>\d+) gen[^/]*/(?P<PresolvedNumConVars>\d+) continuous"
        ),
        re.compile(
            r"Semi-Variable types: (?P<PresolvedNumSemiContVars>\d+) continuous, (?P<PresolvedNumSemiIntVars>\d+) integer$"
        ),
        re.compile(
            r"\s*QMatrix range\s*\[(?P<MinQCCoeff>[^,]+),\s*(?P<MaxQCCoeff>[^\]]+)\]"
        ),
        re.compile(
            r"\s*QLMatrix range\s*\[(?P<MinQCLCoeff>[^,]+),\s*(?P<MaxQCLCoeff>[^\]]+)\]"
        ),
        re.compile(
            r"\s*Matrix range\s*\[(?P<MinCoeff>[^,]+),\s*(?P<MaxCoeff>[^\]]+)\]"
        ),
        re.compile(
            r"\s*QObjective range\s*\[(?P<MinQObjCoeff>[^,]+),\s*(?P<MaxQObjCoeff>[^\]]+)\]"
        ),
        re.compile(
            r"\s*Objective range\s*\[(?P<MinObjCoeff>[^,]+),\s*(?P<MaxObjCoeff>[^\]]+)\]"
        ),
        re.compile(
            r"\s*Bounds range\s*\[(?P<MinBound>[^,]+),\s*(?P<MaxBound>[^\]]+)\]"
        ),
        re.compile(r"\s*RHS range\s*\[(?P<MinRHS>[^,]+),\s*(?P<MaxRHS>[^\]]+)\]"),
        re.compile(r"\s*QRHS range\s*\[(?P<MinQCRHS>[^,]+),\s*(?P<MaxQCRHS>[^\]]+)\]"),
        re.compile(r"Model has (?P<NumQNZs>\d+) quadratic objective terms?"),
        re.compile(r"Model has (?P<NumQConstrs>\d+) quadratic constraints?"),
        re.compile(r"Model has (?P<NumSOS>\d+) SOS constraints?"),
        re.compile(
            r"Model has (?P<NumPWLObjVars>\d+) piecewise-linear objective terms?"
        ),
        re.compile(r"Model has (?P<NumGenConstrs>\d+) general constraints?"),
        re.compile(r"Distributed MIP job count: (?P<DistributedMIPJobs>\d+)"),
        re.compile(r"Concurrent MIP job count: (?P<ConcurrentJobs>\d+)"),
        re.compile(
            r"Concurrent MIP optimizer: (?P<ConcurrentJobs>\d+) concurrent instances \(\d+ threads per instance\)"
        ),
        re.compile(
            r"Presolved: (?P<PresolvedNumConstrs>\d+) (R|r)ows, (?P<PresolvedNumVars>\d+) (C|c)olumns, (?P<PresolvedNumNZs>\d+) (N|n)on(Z|z)ero(e?)s"
        ),
        re.compile(r"Presolved model has (?P<PresolvedNumSOS>\d+) SOS constraint(s)\n"),
        re.compile(
            r"Presolved model has (?P<PresolvedNumQNZs>\d+) quadratic objective terms"
        ),
        re.compile(r"Presolve time: (?P<PresolveTime>[\d\.]+)s"),
    ]

    # Special case: model solved by presolve
    presolve_all_removed = re.compile(r"Presolve: All rows and columns removed")

    def __init__(self):
        """Initialize the Presolve parser.

        The PresolveParser extends beyond the lines associated with the presolved
        model. Specifically, it includes information for all lines appearing between
        the HeaderParser and the NoRelParser or the RelaxationParser.
        """
        self._summary = {}
        self._started = False

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        if not self._started:
            match = PresolveParser.presolve_start_pattern.match(line)
            if match:
                # The start line encodes information that should be stored
                self._started = True
                self._summary.update(typeconvert_groupdict(match))
                return True
            return False

        for pattern in PresolveParser.presolve_intermediate_patterns:
            match = pattern.match(line)
            if match:
                self._summary.update(typeconvert_groupdict(match))
                return True

        match = PresolveParser.presolve_all_removed.match(line)
        if match:
            self._summary.update(
                {
                    "PresolvedNumConstrs": 0,
                    "PresolvedNumVars": 0,
                    "PresolvedNumNZs": 0,
                }
            )
            return True

        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        return self._summary



import glob
import tempfile

import pandas as pd
import pytest
from pandas.api.types import is_integer_dtype
from pandas.testing import assert_frame_equal, assert_series_equal

import grblogtools as glt


@pytest.fixture(scope="module")
def glass4_summary():
    """Summary data from API call."""
    return glt.parse("data/*.log").summary()


@pytest.fixture(scope="module")
def glass4_progress():
    """Progress data from API call."""
    return {
        "norel": glt.parse("data/*.log").progress("norel"),
        "rootlp": glt.parse("data/*.log").progress("rootlp"),
        "nodelog": glt.parse("data/*.log").progress("nodelog"),
    }


@pytest.fixture(scope="module")
def testlog_summary():
    return glt.parse("tests/assets/*.log").summary()


@pytest.fixture(scope="module")
def testlog_progress():
    return {
        "norel": glt.parse("tests/assets/*.log").progress("norel"),
        "rootlp": glt.parse("tests/assets/*.log").progress("rootlp"),
        "nodelog": glt.parse("tests/assets/*.log").progress("nodelog"),
    }


@pytest.fixture(scope="module")
def merged_log():
    with tempfile.NamedTemporaryFile("w") as fp:
        for path in sorted(glob.glob("data/912-glass4-*.log")):
            with open(path) as infile:
                fp.writelines(infile.readlines())
        fp.flush()
        yield fp.name


def test_merged_log(merged_log):
    summary = glt.parse(merged_log).summary()
    result = summary[["Seed", "Runtime", "LogFilePath", "LogNumber"]]
    expected = pd.DataFrame(
        [
            {"Seed": 0, "Runtime": 35.66, "LogFilePath": merged_log, "LogNumber": 1},
            {"Seed": 1, "Runtime": 42.79, "LogFilePath": merged_log, "LogNumber": 2},
            {"Seed": 2, "Runtime": 11.37, "LogFilePath": merged_log, "LogNumber": 3},
        ]
    )
    assert_frame_equal(result, expected)


def test_summary(testlog_summary):
    assert len(testlog_summary) == 7
    assert set(testlog_summary.columns).issuperset(
        {"Status", "ObjVal", "ReadingTime", "RelaxObj"}
    )


def test_progress(testlog_progress):
    assert len(testlog_progress) == 3
    assert len(testlog_progress["norel"]) == 15
    assert set(testlog_progress["norel"].columns).issuperset(
        {"Time", "BestBd", "Incumbent"}
    )
    assert len(testlog_progress["rootlp"]) == 409
    assert set(testlog_progress["rootlp"].columns).issuperset(
        {"Iteration", "PInf", "DInf", "PObj", "DObj"}
    )
    assert len(testlog_progress["nodelog"]) == 133
    assert set(testlog_progress["nodelog"].columns).issuperset(
        {"Depth", "IntInf", "Incumbent", "BestBd", "ItPerNode", "ModelFile", "Version"}
    )


def test_summary_glass4(glass4_summary):
    assert len(glass4_summary) == 63
    assert set(glass4_summary.columns).issuperset(
        {"Status", "ObjVal", "ReadingTime", "RelaxObj", "Seed"}
    )


def test_progress_glass4(glass4_progress):
    assert len(glass4_progress) == 3
    assert len(glass4_progress["norel"]) == 0
    assert len(glass4_progress["rootlp"]) == 0

    assert set(glass4_progress["nodelog"].columns).issuperset(
        {"Depth", "IntInf", "Incumbent", "BestBd", "ItPerNode"}
    )


def test_logfile(glass4_summary):
    logfiles = glass4_summary["LogFilePath"]
    assert len(logfiles.unique()) == len(logfiles)
    assert logfiles.str.startswith("data/").all()
    assert logfiles.str.endswith(".log").all()
    assert_series_equal(
        glass4_summary["LogFile (Parameter)"].apply(lambda l: "data/" + l),
        logfiles,
        check_names=False,
    )
    # log names are stripped of the model name and seed
    log = glass4_summary["Log"]
    assert log.str.startswith("912").all()
    assert not log.str.contains("glass4").any()
    assert not log.str.endswith("-").any()
    assert len(log.unique()) == 21  # different seeds get same label


def test_modelfile(glass4_summary):
    modelfiles = glass4_summary["ModelFilePath"]
    assert len(modelfiles.unique()) == 1
    assert glass4_summary["ModelFile"].eq("glass4").all()
    assert glass4_summary["Model"].eq("glass4").all()


def test_parameters(glass4_summary):
    seeds = glass4_summary["Seed"]
    assert is_integer_dtype(seeds)
    assert seeds.notnull().all()
    assert set(seeds.unique()) == {0, 1, 2}
    assert "TimeLimit (Parameter)" in glass4_summary.columns
    parameters = glass4_summary[
        [f"{c} (Parameter)" for c in ["Heuristics", "MIPFocus", "Presolve"]]
    ]
    assert parameters.notnull().all().all()


def test_legacy_api():
    glass4_summary = glt.get_dataframe(["data/*.log"])
    assert glass4_summary.shape[0] == 63
    assert set(glass4_summary.columns).issuperset({"Status", "ObjVal"})


def test_legacy_api_twopattern():
    """Check multiple file patterns can be passed and duplicate files are
    filtered out."""
    glass4_summary = glt.get_dataframe(
        [
            "data/912-glass4-*.log",
            "data/912-Cuts0-glass4-*.log",
            "data/912-Cuts0-glass4-*.log",
        ]
    )
    assert glass4_summary.shape[0] == 6
    assert set(glass4_summary.columns).issuperset({"Status", "ObjVal"})


def test_listpattern():
    result = glt.parse(
        [
            "data/912-glass4-*.log",
            "data/912-Cuts0-glass4-*.log",
            "data/912-Cuts0-glass4-*.log",
        ]
    )
    glass4_summary = result.summary()
    assert glass4_summary.shape[0] == 6
    assert set(glass4_summary.columns).issuperset({"Status", "ObjVal"})


def test_gurobipy_logs():
    """API logs differ in some cases: sometimes no model file is read, and
    parameter settings may appear before the header. At the moment this test
    just asserts no errors."""
    glt.parse("tests/assets/gurobipy/empty.log").summary()
    glt.parse("tests/assets/gurobipy/read.log").summary()
    glt.parse("tests/assets/gurobipy/*.log").summary()


def test_work():
    """Check if work is correctly parsed for different log files."""
    result = glt.parse("tests/assets/*.log")
    summary = result.summary()

    # Check if work column present
    assert set(summary.columns).issuperset({"Work"})

    # Check if Runtime and Work found
    assert summary["Work"].count() == 6


def test_changed_params():
    result = glt.parse("tests/assets/*.log")
    summary = result.summary()
    assert set(summary.columns).issuperset({"ChangedParams"})
    assert summary["ChangedParams"].apply(lambda d: isinstance(d, dict)).all()
    assert summary["ChangedParams"].count() == 7


def test_create_label():
    # Test expected workflow for creating custom labels

    def label(row):
        params = row["ChangedParams"]
        if params:
            paramstr = "-".join(
                f"{k}={v}" for k, v in sorted(params.items()) if k != "TimeLimit"
            )
        else:
            paramstr = "Default"
        version = row["Version"].replace(".", "")
        return f"{version}-{paramstr}"

    summary = (
        glt.parse("tests/assets/*.log")
        .summary()
        .assign(Label=lambda df: df.apply(label, axis="columns"))
    )
    assert set(summary["Label"]) == {
        "950-CrossoverBasis=1-Method=2",
        "950-Default",
        "950-Method=0",
        "950-Method=2",
        "950-Method=3",
        "950-NonConvex=2",
        "912-NoRelHeurWork=60",
    }



import re

from grblogtools.parsers.util import (
    convert_data_types,
    float_pattern,
    typeconvert_groupdict,
)


class NodeLogParser:
    tree_search_start = re.compile(r" Expl Unexpl(.*)It/Node Time$")

    tree_search_final_stats = [
        re.compile(
            r"Explored (?P<NodeCount>\d+) nodes \((?P<IterCount>\d+) simplex iterations\) in (?P<Runtime>{0}) seconds \((?P<Work>{0}) work units\)".format(
                float_pattern
            )
        ),
        re.compile(
            r"Explored (?P<NodeCount>\d+) nodes \((?P<IterCount>\d+) simplex iterations\) in (?P<Runtime>[^\s]+) seconds"
        ),
        re.compile(
            r"Best objective (?P<ObjVal>[^,]+), best bound (?P<ObjBound>[^,]+), gap (?P<MIPGap>.*)$"
        ),
    ]

    line_types = [
        # tree_search_full_log_line_regex
        re.compile(
            r"\s\s*(?P<CurrentNode>\d+)\s+(?P<RemainingNodes>\d+)\s+(?P<Obj>{0})\s+(?P<Depth>\d+)\s+(?P<IntInf>\d+)\s+(?P<Incumbent>({0}|-))\s+(?P<BestBd>{0})\s+(?P<Gap>(-|{0}%))\s+(?P<ItPerNode>({0}|-))\s+(?P<Time>\d+)s".format(
                float_pattern
            )
        ),
        # tree_search_nodepruned_line_regex
        re.compile(
            r"\s\s*(?P<CurrentNode>\d+)\s+(?P<RemainingNodes>\d+)\s+(?P<Pruned>(cutoff|infeasible|postponed))\s+(?P<Depth>\d+)\s+(?P<Incumbent>(-|{0}))\s+(?P<BestBd>{0})\s+(?P<Gap>(-|{0}%))\s+(?P<ItPerNode>({0}|-))\s+(?P<Time>\d+)s".format(
                float_pattern
            )
        ),
        # tree_search_new_solution_heuristic_log_line_regex
        re.compile(
            r"(?P<NewSolution>H)\s*(?P<CurrentNode>\d+)\s+(?P<RemainingNodes>\d+)\s+(?P<Incumbent>({0}|-))\s+(?P<BestBd>{0})\s+(?P<Gap>{0}%)\s+(?P<ItPerNode>(-|{0}))\s+(?P<Time>\d+)s".format(
                float_pattern
            )
        ),
        # tree_search_new_solution_branching_log_line_regex
        re.compile(
            r"(?P<NewSolution>\*)\s*(?P<CurrentNode>\d+)\s+(?P<RemainingNodes>\d+)\s+(?P<Depth>\d+)\s+(?P<Incumbent>({0}|-))\s+(?P<BestBd>{0})\s+(?P<Gap>{0}%)\s+(?P<ItPerNode>({0}|-))\s+(?P<Time>\d+)s".format(
                float_pattern
            )
        ),
    ]
    cut_report_start = re.compile(r"Cutting planes:")
    cut_report_line = re.compile(r"  (?P<Name>[\w\- ]+): (?P<Count>\d+)")

    def __init__(self):
        """Initialize the NodeLog parser."""
        self._summary = {}
        self._cuts = {}
        self._progress = []
        self._in_cut_report = False
        self._started = False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        summary = self._summary
        summary.update({f"Cuts: {name}": count for name, count in self._cuts.items()})
        return summary

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary and progress data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """

        for regex in self.tree_search_final_stats:
            match = regex.match(line)
            if match:
                entry = typeconvert_groupdict(match)
                self._summary.update(entry)
                return True

        match = self.cut_report_start.match(line)
        if match:
            self._in_cut_report = True
            return True

        if self._in_cut_report:
            match = self.cut_report_line.match(line)
            if match:
                self._cuts[match.group("Name")] = convert_data_types(
                    match.group("Count")
                )
                return True

        # Wait for the header before matching any log lines.
        if not self._started:
            match = self.tree_search_start.match(line)
            if match:
                self._started = True
                return True
            return False

        # Match log lines.
        for regex in self.line_types:
            match = regex.match(line)
            if match:
                self._progress.append(typeconvert_groupdict(match))
                return True

        return False

    def get_progress(self) -> list:
        """Return the progress of the search tree."""
        result = list(self._progress)
        if "Runtime" in self._summary:
            # Final statistics are added as a final tracked line.
            result.append(
                {
                    "Incumbent": self._summary.get("ObjVal"),
                    "BestBd": self._summary.get("ObjBound"),
                    "Gap": self._summary.get("MIPGap"),
                    "CurrentNode": self._summary.get("NodeCount"),
                    "Time": self._summary.get("Runtime"),
                }
            )
        return result



import pytest

import grblogtools as glt


@pytest.fixture
def glass4_results():
    return glt.parse("data/912*glass4*.log")


def test_plot_summary(glass4_results):
    glt.plot(glass4_results.summary(), type="box")


def test_plot_nodelog(glass4_results):
    glt.plot(glass4_results.progress("nodelog"), y="Gap", color="Log", type="line")



import pandas as pd
from pandas.api.types import is_categorical_dtype, is_float_dtype, is_integer_dtype
from pandas.testing import assert_frame_equal

import grblogtools as glt
from grblogtools.helpers import (
    fill_default_parameters,
    fill_default_parameters_nosuffix,
)


def test_parameter_values():
    summary = glt.get_dataframe(["data/*.log"])
    mipfocus = summary["MIPFocus (Parameter)"]
    assert not mipfocus.isnull().any()
    assert is_integer_dtype(mipfocus)
    assert set(mipfocus.unique()) == {0, 1, 2, 3}


def test_pretty_parameters():
    summary = glt.get_dataframe(["data/*.log"], prettyparams=True)
    presolve = summary["Presolve (Parameter)"]
    assert is_categorical_dtype(presolve)
    assert set(presolve.unique()) == {
        "-1: Automatic",
        "1: Conservative",
        "2: Aggressive",
    }
    assert is_float_dtype(summary["Heuristics (Parameter)"])


def test_fill_default_parameters():
    """Test that '* (Parameter)' columns are default-filled using the
    appropriate Version."""
    summary = pd.DataFrame(
        [
            {"Version": "9.1.2", "CrossoverBasis (Parameter)": 1, "Status": "OPTIMAL"},
            {"Version": "9.1.2", "MIPFocus (Parameter)": 1, "Status": "TIME_LIMIT"},
            {"Version": "9.5.0", "Heuristics (Parameter)": 0, "Status": "OPTIMAL"},
        ]
    )
    filled = fill_default_parameters(summary)
    expected = pd.DataFrame(
        [
            {
                "Version": "9.1.2",
                "CrossoverBasis (Parameter)": 1,
                "MIPFocus (Parameter)": 0,
                "Heuristics (Parameter)": 0.05,
                "Status": "OPTIMAL",
            },
            {
                "Version": "9.1.2",
                "CrossoverBasis (Parameter)": 0,
                "MIPFocus (Parameter)": 1,
                "Heuristics (Parameter)": 0.05,
                "Status": "TIME_LIMIT",
            },
            {
                "Version": "9.5.0",
                "CrossoverBasis (Parameter)": -1,
                "MIPFocus (Parameter)": 0,
                "Heuristics (Parameter)": 0.0,
                "Status": "OPTIMAL",
            },
        ]
    )
    assert_frame_equal(filled.sort_index(axis=1), expected.sort_index(axis=1))


def test_fill_default_parameters_nosuffix():
    """The only columns in this input frame are the"""
    parameters = pd.DataFrame(
        [
            {"Version": "9.1.2", "CrossoverBasis": 1},
            {"Version": "9.1.2", "MIPFocus": 1},
            {"Version": "9.5.0", "Heuristics": 0},
        ]
    )
    filled = fill_default_parameters_nosuffix(parameters)
    expected = pd.DataFrame(
        [
            {
                "Version": "9.1.2",
                "CrossoverBasis": 1,
                "MIPFocus": 0,
                "Heuristics": 0.05,
            },
            {
                "Version": "9.1.2",
                "CrossoverBasis": 0,
                "MIPFocus": 1,
                "Heuristics": 0.05,
            },
            {"Version": "9.5.0", "CrossoverBasis": -1, "MIPFocus": 0, "Heuristics": 0},
        ]
    )
    assert_frame_equal(filled.sort_index(axis=1), expected.sort_index(axis=1))



from grblogtools.parsers.continuous import ContinuousParser
from grblogtools.parsers.header import HeaderParser
from grblogtools.parsers.nodelog import NodeLogParser
from grblogtools.parsers.norel import NoRelParser
from grblogtools.parsers.presolve import PresolveParser
from grblogtools.parsers.termination import TerminationParser
from grblogtools.parsers.util import model_type


class SingleLogParser:
    """This class parses one single log file.

    It expects parse to be called once for each line in a log file.
    """

    def __init__(self):
        # Parsers in sequence
        self.header_parser = HeaderParser()
        self.presolve_parser = PresolveParser()
        self.norel_parser = NoRelParser()
        self.continuous_parser = ContinuousParser()
        self.nodelog_parser = NodeLogParser()
        self.termination_parser = TerminationParser()

        # State
        self.started = False
        self.current_parser = self.header_parser
        self.future_parsers = [
            self.presolve_parser,
            self.norel_parser,
            self.continuous_parser,
            self.nodelog_parser,
        ]

    def get_summary(self):
        """Return a summary dict, a merged result of the sub-parser results."""
        summary = {}
        summary.update(self.header_parser.get_summary())
        summary.update(self.presolve_parser.get_summary())
        summary.update(self.norel_parser.get_summary())
        summary.update(self.continuous_parser.get_summary())
        summary.update(self.nodelog_parser.get_summary())
        summary.update(self.termination_parser.get_summary())
        summary["ModelType"] = model_type(
            discrete_vars=summary.get("PresolvedNumBinVars", 0)
            + summary.get("PresolvedNumIntVars", 0)
            + summary.get("PresolvedNumSemiContVars", 0)
            + summary.get("PresolvedNumSemiIntVars", 0),
            quad_nonzeros=summary.get("NumQNZs", 0),
            quad_constrs=summary.get("NumQConstrs", 0),
        )
        summary["ChangedParams"] = self.header_parser.changed_params()
        return summary

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate the component parsers in sequence.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        # Initially, only check the header parser until started
        if not self.started:
            assert self.current_parser is self.header_parser
            matched = self.current_parser.parse(line)
            if matched:
                self.started = True
            return matched

        # First try the current parser
        assert self.current_parser not in self.future_parsers
        if self.current_parser.parse(line):
            return True

        # Check if any future parsers should take over
        for i, parser in enumerate(self.future_parsers):
            if parser.parse(line):
                self.current_parser = parser
                self.future_parsers = self.future_parsers[i + 1 :]
                return True

        # Check if the line matches any pattern of the termination parser
        if self.termination_parser.parse(line):
            return True

        # Nothing matched
        return False



import re

from grblogtools.parsers.util import float_pattern, typeconvert_groupdict


class BarrierParser:
    # The pattern indicating the initialization of the parser
    barrier_start_pattern = re.compile(
        r"Iter(\s+)Primal(\s+)Dual(\s+)Primal(\s+)Dual(\s+)Compl(\s+)Time"
    )

    barrier_ordering_pattern = re.compile(r"Ordering time: (?P<OrderingTime>[\d\.]+)s")

    # The pattern indicating the barrier progress
    barrier_progress_pattern = re.compile(
        r"\s*(?P<Iteration>\d+)(?P<Indicator>\s|\*)\s+(?P<PObj>[^\s]+)\s+(?P<DObj>[^\s]+)\s+(?P<PRes>[^\s]+)\s+(?P<DRes>[^\s]+)\s+(?P<Compl>[^\s]+)\s+(?P<Time>\d+)s"
    )

    # The pattern indicating the crossover
    barrier_crossover_pattern = re.compile(
        r"\s*Push phase complete: Pinf (?P<PushPhasePInf>[^\s]+), Dinf (?P<PushPhaseDInf>[^\s]+)\s+(?P<PushPhaseEndTime>\d+)s"
    )

    # The pattern indicating the termination of the barrier algorithm
    barrier_termination_patterns = [
        re.compile(
            r"Barrier solved model in (?P<BarIterCount>[^\s]+) iterations and (?P<Runtime>{0}) seconds \((?P<Work>{0}) work units\)".format(
                float_pattern
            )
        ),
        re.compile(
            r"Barrier performed (?P<BarIterCount>\d+) iterations in (?P<Runtime>{0}) seconds \((?P<Work>{0}) work units\)".format(
                float_pattern
            )
        ),
        re.compile(
            r"Barrier solved model in (?P<BarIterCount>[^\s]+) iterations and (?P<Runtime>[^\s]+) seconds"
        ),
        re.compile(
            r"Barrier performed (?P<BarIterCount>\d+) iterations in (?P<Runtime>[^\s]+) seconds"
        ),
    ]

    def __init__(self):
        """Initialize the Barrier parser."""
        self._summary = {}
        self._progress = []
        self._started = False

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary and progress data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        barrier_ordering_match = BarrierParser.barrier_ordering_pattern.match(line)
        if barrier_ordering_match:
            self._summary.update(typeconvert_groupdict(barrier_ordering_match))
            return True

        if not self._started:
            match = BarrierParser.barrier_start_pattern.match(line)
            if match:
                self._started = True
                return True
            return False

        progress_match = BarrierParser.barrier_progress_pattern.match(line)
        if progress_match:
            entry = {"Type": "barrier"}
            entry.update(typeconvert_groupdict(progress_match))
            self._progress.append(entry)
            return True

        for barrier_termination_pattern in BarrierParser.barrier_termination_patterns:
            barrier_termination_match = barrier_termination_pattern.match(line)
            if barrier_termination_match:
                self._summary.update(typeconvert_groupdict(barrier_termination_match))
                return True

        crossover_match = BarrierParser.barrier_crossover_pattern.match(line)
        if crossover_match:
            self._summary.update(typeconvert_groupdict(crossover_match))
            return True

        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        return self._summary

    def get_progress(self) -> list:
        """Return the detailed progress in the barrier method."""
        return self._progress



import re
from functools import partial
from pathlib import Path

from grblogtools.parameters import load_defaults, load_descriptions

re_parameter_column = re.compile(r"(.*) \(Parameter\)")


def fill_for_version(group, parameter_columns):
    parameter_defaults = load_defaults(
        version=group["Version"].iloc[0].replace(".", "")
    )
    for column in parameter_columns:
        default = parameter_defaults.get(re_parameter_column.match(column).group(1))
        if default is not None:
            group[column] = group[column].fillna(default).astype(type(default))
    return group


def fill_default_parameters(summary):
    """Fill NaN parameter values with the actual default value."""
    parameter_columns = [
        column
        for column, series in summary.items()
        if re_parameter_column.match(column) and series.isnull().any()
    ]
    # TODO test cases where there are different versions involved
    return summary.groupby("Version", group_keys=False).apply(
        partial(fill_for_version, parameter_columns=parameter_columns)
    )


def fill_for_version_nosuffix(group):
    parameter_defaults = load_defaults(
        version=group["Version"].iloc[0].replace(".", "")
    )
    for parameter in group.columns:
        default = parameter_defaults.get(parameter)
        if default is not None:
            group[parameter] = group[parameter].fillna(default).astype(type(default))
    return group


def fill_default_parameters_nosuffix(parameters):
    """Fill defaults for Version and parameter cols with no (Parameter) suffix."""
    return parameters.groupby("Version", group_keys=False).apply(
        fill_for_version_nosuffix
    )


def add_categorical_descriptions(summary):
    """Replace some columns with categorical descriptions if available.

    It modifies the summary dict in place.
    """
    descriptions = load_descriptions()
    parameter_columns = [column for column in summary.columns if column in descriptions]
    for column in parameter_columns:
        summary[column] = summary[column].map(descriptions[column]).astype("category")
    return summary


def strip_model_and_seed(row):
    """Return the Log name.

    If the log path contains the model name, return everything to the left.
    Otherwise, just return the log stem.

    i.e. with Model = 'glass4'
        data/912-Cuts0-glass4-0.log -> 912-Cuts0
        data/some-log.log -> some-log
    """
    if row["Model"] is None:
        return None
    log_stem = Path(row["LogFilePath"]).stem
    run, mid, _ = log_stem.partition(row["Model"])
    if mid and run:
        return run.rstrip("-")
    return log_stem
